/**
* OLAT - Online Learning and Training<br>
* http://www.olat.org
* <p>
* Licensed under the Apache License, Version 2.0 (the "License"); <br>
* you may not use this file except in compliance with the License.<br>
* You may obtain a copy of the License at
* <p>
* http://www.apache.org/licenses/LICENSE-2.0
* <p>
* Unless required by applicable law or agreed to in writing,<br>
* software distributed under the License is distributed on an "AS IS" BASIS, <br>
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. <br>
* See the License for the specific language governing permissions and <br>
* limitations under the License.
* <p>
* Copyright (c) since 2004 at Multimedia- & E-Learning Services (MELS),<br>
* University of Zurich, Switzerland.
* <hr>
* <a href="http://www.openolat.org">
* OpenOLAT - Online Learning and Training</a><br>
* This file has been modified by the OpenOLAT community. Changes are licensed
* under the Apache 2.0 license as the original file.  
* <p>
*/ 

package org.olat.core.commons.persistence;

import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;

import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.ejb.HibernateEntityManager;
import org.hibernate.type.Type;
import org.olat.core.commons.persistence.DBImpl.ThreadLocalData;
import org.olat.core.logging.AssertException;
import org.olat.core.logging.DBRuntimeException;
import org.olat.core.manager.BasicManager;

/**
 * 
 * @author Andreas Ch. Kapp
 * @author Christian Guretzki
 */
class DBManager extends BasicManager {

	DBManager() {
		//
	}
	
	private boolean unusableTrx(EntityTransaction trx) {
		return trx == null || !trx.isActive() || trx.getRollbackOnly();
	}

	/**
	 * Hibernate Update method. Use this in a transactional context by using
	 * your current transaction object.
	 * @param trx The current db transaction
	 * @param o Object to be updated
	 */	
	void updateObject(EntityManager em, Object o, ThreadLocalData data) {
		EntityTransaction trx = em.getTransaction();
		if (unusableTrx(trx)) { // some program bug
			throw new DBRuntimeException("cannot update in a transaction that is rolledback or committed " + o);
		}
		try {
			getSession(em).update(o);
			if (isLogDebugEnabled()) {
				logDebug("update (trans "+trx.hashCode()+") class "+o.getClass().getName()+" = "+o.toString());	
			}									
		} catch (HibernateException e) { // we have some error
			trx.setRollbackOnly();
			data.setError(e);
			throw new DBRuntimeException("Update object failed in transaction. Query: " +  o , e);
		}
	}
	
	/**
	 * Hibernate Delete method. Use this in a transactional context by using
	 * your current transaction object.
	 * @param trx The current db transaction
	 * @param o Object to be deleted
	 */	
	void deleteObject(EntityManager em,  Object o, ThreadLocalData data) {
		EntityTransaction trx = em.getTransaction();
		if (unusableTrx(trx)) { // some program bug
			throw new DBRuntimeException("cannot delete in a transaction that is rolledback or committed " + o);
		}
		try {
			Object relaoded = em.merge(o);
			em.remove(relaoded);
			if (isLogDebugEnabled()) {
				logDebug("delete (trans "+trx.hashCode()+") class "+o.getClass().getName()+" = "+o.toString());	
			}
		} catch (HibernateException e) { // we have some error
			trx.setRollbackOnly();
			data.setError(e);
			throw new DBRuntimeException("Delete of object failed: " + o, e);
		}
	}
	
	/**
	 * Hibernate Save method. Use this in a transactional context by using
	 * your current transaction object.
	 * @param trx The current db transaction
	 * @param o Object to be saved
	 */	
	void saveObject(EntityManager em, Object o, ThreadLocalData data) {
		EntityTransaction trx = em.getTransaction();
		if (unusableTrx(trx)) { // some program bug
			throw new DBRuntimeException("cannot save in a transaction that is rolledback or committed: " + o);
		}
		try {
			em.persist(o);
			if (isLogDebugEnabled()) {
				logDebug("save (trans "+trx.hashCode()+") class "+o.getClass().getName()+" = "+o.toString());	
			}						
		} catch (Exception e) { // we have some error
			trx.setRollbackOnly();
			data.setError(e);
			throw new DBRuntimeException("Save failed in transaction. object: " +  o, e);
		}
	}

	/**
	 * Hibernate Load method. Use this in a transactional context by using
	 * your current transaction object.
	 * @param trx The current db transaction
	 * @param theClass The class for the object to be loaded
	 * @param pK The primary key for the object
	 */	
	<U> U loadObject(EntityManager em, Class<U> theClass, Long pK) {
		U o = null;
		EntityTransaction trx = em.getTransaction();
		try {
			o = em.find(theClass, pK);
			if (isLogDebugEnabled()) {
				logDebug("load (res " +(o == null? "null": "ok")+")(trans "+trx.hashCode()+") key "+pK+" class "+theClass.getName());	
			}
		} catch (Exception e) {
			String msg = "loadObject error: " + theClass + " " + pK + " ";
			throw new DBRuntimeException(msg, e);
		}
		return o;
	}
	
	/**
	 * Hibernate Find method. Use this in a transactional context by using
	 * your current transaction object.
	 * @param trx The current db transaction
	 * @param query The HQL query
	 * @param value The value to search for
	 * @param type The Hibernate datatype of the search value 
	 */	
	List find(EntityManager em, String query, Object value, Type type, ThreadLocalData data) {
		List li = null;
		EntityTransaction trx = em.getTransaction();
		try {
			boolean doLog = isLogDebugEnabled();
			long start = 0;
			if (doLog) start = System.currentTimeMillis();

			// old: li = this.getSession().find(query, value, type);
			Query qu = getSession(em).createQuery(query);
			qu.setParameter(0, value, type);
			li = qu.list();
			
			if (doLog) {
				long time = (System.currentTimeMillis() - start);
				logQuery("find (time "+time+", res " +(li == null? "null": ""+li.size())+")(trans "+trx.hashCode()+")", new Object[] {value}, new Type[] {type}, query);	
			}
		} catch (HibernateException e) {
			trx.setRollbackOnly();
			String msg = "Find failed in transaction. Query: " +  query + " " + e;
			data.setError(e);
			throw new DBRuntimeException(msg, e);
		}
		return li;
	}

	/**
	 * Hibernate Find method. Use this in a transactional context by using
	 * your current transaction object.
	 * @param trx The current db transaction
	 * @param query The HQL query
	 * @param values The object array containing all search values 
	 * @param types The object array containing all Hibernate datatype of the search values 
	 */	
	List find(EntityManager em, String query, Object [] values, Type [] types, ThreadLocalData data) {
		List li = null;
		EntityTransaction trx = em.getTransaction();
		try {
			boolean doLog = isLogDebugEnabled();
			long start = 0;
			if (doLog) start = System.currentTimeMillis();
			// old: li = getSession().find(query, values, types);
			Query qu = getSession(em).createQuery(query);
			qu.setParameters(values, types);
			li = qu.list();
			
			
			if (doLog) {
				long time = (System.currentTimeMillis() - start);
				logQuery("find (time "+time+", res " +(li == null? "null": ""+li.size())+")(trans "+trx.hashCode()+")", values, types, query);	
			}			
	
		} catch (HibernateException e) {
			trx.setRollbackOnly();
			String msg = "Find failed in transaction. Query: " +  query + " " + e;
			data.setError(e);
			throw new DBRuntimeException(msg, e);
		}
		return li;
	}
	
	/**
	 * Hibernate find method. Use this in a transactional context by using
	 * your current transaction object.
	 * @param trx The current db transaction
	 * @param query The HQL query
	 */	
	List find(EntityManager em, String query, ThreadLocalData data) {
		List li = null;
		EntityTransaction trx = em.getTransaction();
		try {
			boolean doLog = isLogDebugEnabled();
			long start = 0;
			if (doLog) start = System.currentTimeMillis();
			// old: li = getSession().find(query);
			Query qu = getSession(em).createQuery(query);
			li = qu.list();

			if (doLog) {
				long time = (System.currentTimeMillis() - start);
				logQuery("find (time "+time+", res " +(li == null? "null": ""+li.size())+")(trans "+trx.hashCode()+")", null, null, query);	
			}
		} catch (HibernateException e) {
			String msg = "Find in transaction failed: " + query + " " + e;
			trx.setRollbackOnly();
			data.setError(e);
			throw new DBRuntimeException(msg, e);
		}
		return li;
	}

	
	Session getSession(EntityManager em) {
		return em.unwrap(HibernateEntityManager.class).getSession();
	}
	
	/**
	 * @param query
	 * @return Hibernate Query object.
	 */
	DBQuery createQuery(EntityManager em, String query, ThreadLocalData data) {
		Query q = null;
		DBQuery dbq = null;
		try {
			q = getSession(em).createQuery(query);
			dbq = new DBQueryImpl(q);
		} catch (HibernateException he) {
			data.setError(he);
			throw new DBRuntimeException("Error while creating DBQueryImpl: ", he);
		}
		return dbq;
	}
	
	DBQuery createNamedQuery(EntityManager em, String queryName, String dbVendor, boolean vendorSpecific) {
		if (queryName == null) {
			throw new AssertException("queryName must not be NULL");
		}
		Query q= null;
		DBQuery dbq = null;
		if (vendorSpecific) {
			String finalQueryName = vendorSpecific ? dbVendor + "_" + queryName : queryName;
			q = getSession(em).getNamedQuery(finalQueryName);
			if (q == null) { 
				// try fallback with normal query
				q = getSession(em).getNamedQuery(queryName);
			}
			if (q == null) {
				String msg = "Can not create namedQuery::" + finalQueryName;
				if (vendorSpecific) {
					msg += " for dbvendor::" + dbVendor + " and non db specific::" + queryName;
				}
				msg += ", named query does not exist";
				logError(msg, null);
			}	else {

				dbq = new DBQueryImpl(q);
			}
		}
		return dbq;
	}

	/**
	 * Hibernate Contains method
	 * @param object
	 * @return True if the Object instance is in the Hibernate Session.
	 */
	boolean contains(EntityManager em, Object object) {
		return em.contains(object);
	}
	

	/**
	 * Hibernates Evict method
	 * @param object The object to be removed from hibernates session cache
	 */
	void evict(EntityManager em, Object object, ThreadLocalData data) {
		try {
			getSession(em).evict(object);			
		} catch (Exception e) {
			data.setError(e);
			throw new DBRuntimeException("Error in evict() Object from Database. ", e);
		}
	}
	
	/**
	 * Hibernate refresh method
	 * @return object a persistent or detached instance
	 */
	void refresh(EntityManager em, Object object, ThreadLocalData data) {
		try{
			em.refresh(object);
		} catch(Exception e) {
			data.setError(e);
			throw new DBRuntimeException("Error in refresh() Object from Database. ", e);
		}
	}
	
	int delete(EntityManager em, String query, Object value, Type type) {	
		int deleted = 0;

		EntityTransaction trx = em.getTransaction();
		if (unusableTrx(trx)) { // some program bug
			throw new DBRuntimeException("cannot delete in a transaction that is rolledback or committed " + value);
		}
		try {
			//old: deleted = getSession().delete(query, value, type);
			
			Session si = getSession(em);
			Query qu = si.createQuery(query);
			qu.setParameter(0, value, type);
			List foundToDel = qu.list();
			int deletionCount = foundToDel.size();
			for (int i = 0; i < deletionCount; i++ ) {
				si.delete( foundToDel.get(i) );
			}
			
			if (isLogDebugEnabled()) {
				logQuery("delete (trans "+trx.hashCode()+")",new Object[] {value}, new Type[] {type}, query);	
			}
		} catch (HibernateException e) { // we have some error
			trx.setRollbackOnly();
			throw new DBRuntimeException ("Could not delete object: " + value, e);
		}
		return deleted;
	}

	int delete(EntityManager em, String query, Object[] values, Type[] types) {
		int deleted = 0;
		EntityTransaction trx = em.getTransaction();
		if (unusableTrx(trx)) { // some program bug
			throw new DBRuntimeException("cannot delete in a transaction that is rolledback or committed " + values);
		}
		try {
			//old: deleted = getSession().delete(query, values, types);
			Session si = getSession(em);
			Query qu = si.createQuery(query);
			qu.setParameters(values, types);
			List foundToDel = qu.list();
			deleted = foundToDel.size();
			for (int i = 0; i < deleted; i++ ) {
				si.delete( foundToDel.get(i) );
			}
			
			if (isLogDebugEnabled()) {
				logQuery("delete (trans "+trx.hashCode()+")", values, types, query);	
			}			
		} catch (HibernateException e) { // we have some error
			trx.setRollbackOnly();
			throw new DBRuntimeException ("Could not delete object: " + values, e);
		}
		return deleted;
	}

	/**
	 * Find a Object by its identifier. 
	 * This method should be used instead of loadObject when it is ok to not find the object.
	 * In opposite to the loadObject method this method will return null and not throw a 
	 * DBRuntimeException when the object can't be loaded.
	 * @param theClass
	 * @param key
	 * @return an persistent object.
	 */
	Object findObject(EntityManager em, Class theClass, Long key) {
		//o_clusterREVIEW see Session.java
		Object o = getSession(em).get(theClass, key);
		if (isLogDebugEnabled()) {
			logDebug("findload (res " +(o == null? "null": "ok")+") key "+key+" class "+theClass.getName());	
		}			
		return o;
	}
	
	private void logQuery(String info, Object[] values, Type[] types, String query) {
		StringBuilder sb = new StringBuilder(info);
		sb.append(": args: ");
		if (values == null) {
			sb.append(" none");
		}
		else {
			for (int i = 0; i < values.length; i++) {
				Object val = values[i];
				Type type = types[i];
				sb.append((val == null? "NULL": val.toString())).append("(").append(type.getName()).append(" ), ");
			}
		}
		sb.append(", query: ").append(query);
		logDebug(sb.toString());
	}
}